#if HAVE_CONFIG_H
#include <ldas_tools_config.h>
#endif /* HAVE_CONFIG_H */

#include <iomanip>
#include <memory>
#include <sstream>

#include "ldastoolsal/fstream.hh"
#include "ldastoolsal/AtExit.hh"
#include "ldastoolsal/unittest.h"

#include "framecpp/Common/FrameH.hh"

#include "framecpp/Version8/FrameH.hh"

#include "framecpp/FrameCPP.hh"
#include "framecpp/FrVect.hh"
#include "framecpp/OFrameStream.hh"

#if 0
using std::filebuf;
using std::ofstream;
#else
using LDASTools::AL::filebuf;
using LDASTools::AL::ofstream;
#endif

using namespace FrameCPP;

using LDASTools::AL::AutoArray;
using LDASTools::AL::SharedPtr;

typedef Common::FrameBuffer< filebuf > FrameBuffer;
typedef Common::FrameH frameh_type;
typedef SharedPtr< frameh_type > build_frame_type;

static const INT_4U CHANNEL_COUNT = /* 64 */  32;
static const LDASTools::AL::GPSTime START_TIME( 913423140, 0 );
static const INT_2U ULEAPS = START_TIME.GetLeapSeconds( );
static const REAL_8 DT = 60.0;

LDASTools::Testing::UnitTest 	Test;

static build_frame_type build_frame( version_type Version,
					 INT_4U ChannelCount );
void   make_file( INT_4U ChannelCount );

int
main( int argc, char* argv[] )
{
  FrameCPP::Initialize( );

  Test.Init( argc, argv );

  try
  {
    int power = 0;

    for ( INT_4U
	    count = 1;
	  count <= 64;
	  count = 1 << ++power )
    {
      make_file( count );
    }
  }
  catch ( const std::exception& E )
  {
    Test.Check( false ) << "Caught an exception: " << E.what( ) << std::endl;
  }
  catch ( ... )
  {
    Test.Check( false ) << "Caught an unknown exception" << std::endl;
  }
  try
  {
    LDASTools::AL::AtExit::Cleanup( );
  }
  catch( ... )
  {
  }
  Test.Exit();
}

void
make_file( INT_4U ChannelCount )
{
  try
  {
    int frame_spec = FRAME_SPEC_CURRENT;

    LDASTools::AL::GPSTime	start;
    LDASTools::AL::GPSTime	finish;

    start.Now( );
    INT_4U	OBUFFER_SIZE = ( 10 * 1024 * 1024 );
    AutoArray< FrameBuffer::char_type  >
      obuffer( new FrameBuffer::char_type[ OBUFFER_SIZE ] );

    //-------------------------------------------------------------------
    // Construct a frame
    //-------------------------------------------------------------------
    SharedPtr< frameh_type >
      frameh( build_frame( frame_spec, ChannelCount ) );

    finish.Now( );
    Test.Message( 10 )
      << "Time to create frame with " << ChannelCount << "k channels:"
      << " [dt: " << ( finish - start )
      << " Percent of Real Time: " << ( ( finish - start ) * 100 / DT )
      << "]"
      << std::endl
      ;

    //-------------------------------------------------------------------
    // Constuct output frame name
    //-------------------------------------------------------------------
    std::ostringstream	filename;
    filename << "Z-SixtyK_" << ChannelCount
	     << "-" << START_TIME.GetSeconds( )
	     << "-" << int( DT )
	     << ".gwf";
    //-------------------------------------------------------------------
    // Time writing of the frame
    //-------------------------------------------------------------------
    start.Now( );

    std::unique_ptr< FrameBuffer >
      obuf( new FrameBuffer( std::ios::out ) );

    obuf->pubsetbuf( obuffer.get( ), OBUFFER_SIZE );
    obuf->open( filename.str( ).c_str( ),
		std::ios::out | std::ios::binary );

    OFrameStream	ofs( obuf.release( ) );

    ofs.DisableTOC( );

    ofs.WriteFrame( frameh,
		    FrameCPP::FrVect::ZERO_SUPPRESS_OTHERWISE_GZIP,
		    FrameCPP::FrVect::DEFAULT_GZIP_LEVEL );
    ofs.Close( );

    finish.Now( );
    Test.Check( ( finish - start ) < DT )
      << "Writing of frame with " << ChannelCount << "k channels completed in less than real time"
      << "( dt: " << DT
      << " vs write time: " << ( finish - start )
      << " [ " << ( ( finish - start ) * 100 / DT ) << "% ])"
      << std::endl
      ;
    Test.Message( 10 )
      << "Time to write file: " << ( finish - start )
      << " Percent of Real Time: " << ( ( finish - start ) * 100 / DT )
      << std::endl
      ;
  }
  catch ( const std::exception& E )
  {
    Test.Check( false ) << "Caught an exception: " << E.what( ) << std::endl;
  }
  catch ( ... )
  {
    Test.Check( false ) << "Caught an unknown exception" << std::endl;
  }
}

//=======================================================================
// Local functions
//=======================================================================

static void add_channels( version_type Version,
			  build_frame_type Frame,
			  INT_2U padding,
			  INT_4U ChannelCount );

static build_frame_type create_frameh( version_type Version );

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
template< typename RawData_type > void
add_channel( RawData_type RawData, INT_2U Padding, INT_4U ChannelOffset );

template<>
void
add_channel
( FrameCPP::Version_8::FrameH::rawData_type RawData,
  INT_2U Padding,
  INT_4U ChannelOffset )
{
  using namespace FrameCPP::Version_8;

  std::ostringstream	name;
  name << "Z1:Channel"
       << std::setw( Padding ) << std::setfill( '0' ) << ChannelOffset
    ;
  std::vector< REAL_4 > d( size_t( DT ), 0.0 );

  Dimension	dim( d.size( ) );
  Version_8::FrVect	v( name.str( ), 1,
			   &dim, &d[0] );
  FrAdcData	a( name.str( ),	/* name */
		   0, ChannelOffset, /* group, channel */
		   32,		/* nBits */
		   1,		/* sampleRate */
		   0,		/* bias */
		   1,		/* slope */
		   "NONE",	/* units */
		   0,		/* fShift */
		   0,		/* timeOffset */
		   0,		/* dataValid */
		   0		/* phase */
		   );
  // a.RefData( ).append( &v, false /* allocate */, false /* own */ );
  a.RefData( ).append( v );
  RawData->RefFirstAdc( ).append( a );
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
template< typename RawData_type >
void add_adc_channels( RawData_type& RawData, INT_2U Padding, INT_4U ChannelCount )
{
  for ( INT_4U cc = 0;
	cc < ChannelCount;
	++cc )
  {
    add_channel( RawData, Padding, cc );
  }
  
}
//-----------------------------------------------------------------------
// Creation of the header
//-----------------------------------------------------------------------
template< int > build_frame_type frameh( );

template<> build_frame_type
frameh< 8 >( )
{
  using namespace FrameCPP::Version_8;
  using FrameCPP::Version_8::FrameH;

  SharedPtr< FrameH >
    frameh( new FrameH( "SixtyK",
			0,
			0,
			START_TIME,
			ULEAPS,
			DT ) );
  SharedPtr< FrRawData >	r( new FrRawData( "SixtyK" ) );
  frameh->SetRawData( r );
  return LDASTools::AL::DynamicPointerCast< frameh_type >( frameh );
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static void
add_channels( version_type Version,
	      build_frame_type Frame,
	      INT_2U Padding,
	      INT_4U ChannelCount )
{
  switch( Version )
  {
  case 8:
    {
      using namespace FrameCPP::Version_8;
      using FrameCPP::Version_8::FrameH;

      FrameH*	f = dynamic_cast< FrameH* >( Frame.get( ) );
      if ( f )
      {
	FrameH::rawData_type r = f->GetRawData( );
	if ( r )
	{
	  add_adc_channels( r, Padding, ChannelCount);
	}
      }
    }
    break;
  default:
    break;
  }
}

static build_frame_type
build_frame( version_type Version,
	     INT_4U ChannelCount )
{
  build_frame_type	retval;

  retval = create_frameh( Version );

  const INT_4U cc_last = ChannelCount * 1024;
  int padding = 0;

  for ( INT_4U cc_cur = cc_last;
	cc_cur > 10;
	cc_cur /= 10 )
  {
    ++padding;
  }

  add_channels( Version, retval, padding, cc_last );
  return retval;
}

static build_frame_type
create_frameh( version_type Version )
{
  build_frame_type	retval;
  switch( Version )
  {
  case 8:
    retval = frameh< 8 >( );
    break;
  default:
    {
      std::ostringstream	msg;

      msg << "create_frameh does not support version "
	  << (int)Version
	  << " of the frame specification"
	;
      throw std::range_error( msg.str( ) );
    }
    break;
  }
  return retval;
}
