#! /usr/bin/env python
#------------------------------------------------------------------------
# This is a sample script based on user request.
#------------------------------------------------------------------------
from __future__ import division

import frameCPP
import numpy as np

__author__ = "Nickolas Fotopoulos <nickolas.fotopoulos@ligo.org>"


#------------------------------------------------------------------------
# Routine to dump the contents of the FrameH structure
#------------------------------------------------------------------------
def dumpFrameH( frameh ):
    print "FrameH:       name: ", frameh.name
    print "               run: ", frameh.run
    print "             frame: ", frameh.frame
    print "       dataQuality: ", frameh.dataQuality
    print "            ULeapS: ", frameh.ULeapS
    print "                dt: ", frameh.dt
    return

#
# Functions to navigate the filestream
# XXX: Use once reference counting bug is fixed.
#
def iter_vect(stream):
    "Return an iterator of FrVect objects found in stream."
    toc = stream.GetTOC()

    # grab vectors from ProcData
    for channelName in toc.GetProc( ):
        proc = stream.ReadFrProcData(0, channelName )  # what is this 0??
        for v in proc.data:
            yield v

    # grab vectors from AdcData
    toc_adc = toc.GetADC()
    for key in sorted(toc_adc.iterkeys()):
        adc = stream.ReadFrAdcData(0, key)  # what is this 0??
        for v in adc.data:
            yield v

#
# This is an example of how to read FrEvents
#
def frgetevents(filename, name):
    fs = frameCPP.IFrameFStream(filename)
    toc = fs.GetTOC( )
    print "NFrame: ", toc.GetNFrame( )
    for frame_index in range( toc.GetNFrame( ) ):
        event = fs.ReadFrEvent( frame_index, name )
        print "FrEvent.name: ", event.name
        print "FrEvent.comment", event.comment
        print "FrEvent.inputs", event.inputs
        print "FrEvent.GTime", event.GTime
        print "FrEvent.timeBefore", event.timeBefore
        print "FrEvent.timeAfter", event.timeAfter
        print "FrEvent.eventStatus", event.eventStatus
        print "FrEvent.amplitude", event.amplitude
        print "FrEvent.probability", event.probability
        print "FrEvent.statistics", event.statistics
        print "FrEvent.params", event.params
        
#
# The high-level functions that end-users will call
#
def frgetvect1d(filename, channel, method, start=None, end=None):
    fs = frameCPP.IFrameFStream(filename)

    #---------------------------------------------------------------------
    # Show what is being done
    #---------------------------------------------------------------------
    print "META_DATA: filename: ", filename
    print "META_DATA:  channel: ", channel
    print "META_DATA:   method: ", method
    print "META_DATA:    start: ", start
    print "META_DATA:      end: ", end

    #---------------------------------------------------------------------
    # Print information about the frame stream
    #---------------------------------------------------------------------
    print "Frame Specification: ", fs.Version( )
    print "Frame Library: ", fs.FrameLibrary( ), " (", fs.FrameLibraryName( ), ")"
    print "Frame Minor: ", fs.LibraryRevision( )

    #---------------------------------------------------------------------
    # Print information about the table of contents
    #---------------------------------------------------------------------
    print "TOC: FrAdcData Info: ", fs.GetTOC( ).GetADC( )
    print "TOC: FrProcData Info: ", fs.GetTOC( ).GetProc( ), \
          " Length: ", len( fs.GetTOC( ).GetProc( ) )

    #---------------------------------------------------------------------
    #
    #---------------------------------------------------------------------
    sev_list = []  # each row will be (start, end, vector)

    if method == 1:
        #----------------------------------------------------------------
        # This shows how read multiple channels
        #----------------------------------------------------------------
        # XXX: hold references in the local scope to proc and adc structures due to refcount bug;
        #      replace this block with iter_vect eventually
        vects = []
        toc = fs.GetTOC()
        procs = []
        for channelName in toc.GetProc( ):
            # Read the channel toc_proc.Name(i) from frame 0 (zero) as
            #   indexed from the beginning of the file.
            proc = fs.ReadFrProcData(0, channelName )
            procs.append(proc)
            for v in proc.data:
                vects.append(v)

        toc_adc = toc.GetADC()
        adcs = []
        for key in sorted(toc_adc.iterkeys()):
            adc = fs.ReadFrAdcData(0, key)  # what is this 0??
            adcs.append(adc)
            for v in adc.data:
                vects.append(v)
        # end XXX
    elif method == 2:
        #----------------------------------------------------------------
        # This shows how read a single channel
        #----------------------------------------------------------------
        vects = []
        procs = []
        adcs  = []
        adc = fs.ReadFrAdcData( 0, channel )
        if adc:
            adcs.append( adc )
            for v in adc.data:
                vects.append(v)
        else:
            proc = fs.ReadFrProcData( 0, channel )
            if proc:
                procs.append( proc )
                for v in proc.data:
                    vects.append(v)
            # end if - proc
        # end if - adc
    elif method == 3:
        #----------------------------------------------------------------
        # This shows how read a single channel without preserving
        #----------------------------------------------------------------
        vects = []
        adc = fs.ReadFrAdcData( 0, channel )
        if adc:
            for v in adc.data:
                vects.append(v)
        else:
            proc = fs.ReadFrProcData( 0, channel )
            if proc:
                for v in proc.data:
                    vects.append(v)
            # end if - proc
        # end if - adc
    elif method == 4:
        vects = list(iter_vect(fs))
    else:
        print "Unknown method: ", method
        assert( False )

    # check for channel separately for extra debugging help to user
    vects = [vect for vect in vects if vect.name == channel]
    if len(vects) == 0:
        raise ValueError, "In file %s, vector %s not found." % (filename, channel)

    # now filter to vectors that overlap the requested region
    for vect in vects:
        if vect.nDim != 1:
            raise NotImplementedError, "Multiple dimensions are not yet supported."
        print "DEBUG: sizeof dims: ", vect.GetDim(0).nx
        dim = vect.GetDim(0)
        startX = dim.startX
        vect_end = startX + dim.dx * dim.nx
        if (end is None or startX < end) and (start is None or vect_end > start):
            sev_list.append((startX, vect_end, vect))
    if len(sev_list) == 0:
        raise ValueError, "In file %s, vector %s not found between %d and %d." \
            % (filename, channel, start or "-infinity", end or "infinity")
    sev_list.sort()  # sorts by time first

    # extract useful quantities
    vects = [vect for _, _, vect in sev_list]
    dim = vects[0].GetDim(0)
    dx = dim.dx
    true_start = int(round((start - dim.startX) / dx)) * dx + dim.startX
    true_end = int(round((end - vects[-1].GetDim(0).startX) / dx) * dx + vects[-1].GetDim(0).startX) / dx
    expected_num_samples = int(round((true_end - true_start) / dx))

    # sanity checks
    if not all(vect.GetDim(0).dx == dx for vect in vects):
        raise ValueError, "differing sample rates detected"
    for (_, old_end, _), (new_start, _, _) in zip(sev_list, sev_list[1:]):
        if old_end != new_start:
            raise ValueError, "The vectors do not form a contiguous block. "\
                "There is a gap between %f and %f." % (old_end, new_start)

    # extract data
    # NB: extract first array separately to get dtype for the output array
    if start is None: start = sev_list[0][0]
    if end is None: end = sev_list[-1][1]

    now = start
    offset = 0
    v_start, v_end, vect = sev_list[0]
    first_array = vect.GetDataArray()
    start_ind = int(round((now - v_start) / dx))
    end_ind = int(round((min(v_end, end) - v_start) / dx))
    out_data = np.zeros(expected_num_samples, dtype=first_array.dtype)
    out_data[:end_ind - start_ind + 1] = first_array[start_ind:end_ind]
    del first_array
    now = min(v_end, end)
    offset += end_ind - start_ind
    for v_start, v_end, vect in sev_list[1:]:
        start_ind = int(round((now - v_start) / dx))
        assert start_ind >= 0, "gap detected"
        end_ind = int(round((min(v_end, end) - v_start) / dx))
        out_data[offset:offset + end_ind - start_ind + 1] = vect.GetDataArray()[start_ind:end_ind]
        now = min(v_end, end)
        offset += end_ind - start_ind
    assert offset == expected_num_samples

    return (out_data, true_start, 0.0, dim.dx, dim.unitX, vect.unitY)

def frameIteration( filename ):
    #--------------------------------------------------------------------
    # Example of how to read the frames sequentially from a file
    #--------------------------------------------------------------------
    # * Open the frame file
    #--------------------------------------------------------------------
    fs = frameCPP.IFrameFStream(filename)
    #--------------------------------------------------------------------
    # * Loop over the frames
    #--------------------------------------------------------------------
    # frameh = fs.ReadNextFrame( )
    frameh = fs.ReadFrameN( 0 )
    dumpFrameH( frameh )
    return

if __name__ == "__main__":
    for i in range( 1, 5 ):
        # Try a ProcData
        #vect = frgetvect1d("test.gwf", "fastProc")
        #print len(vect[0])
        #print vect

        # Try an AdcData
        #print frgetvect1d("test.gwf", "fastAdc0")

        # Try something made from frameCPP
        frgetevents("Z-R_std_test_frame_ver8-600000000-1.gwf", 'TesT' )
        vect = frgetvect1d("Z-R_std_test_frame_ver8-600000000-1.gwf", \
                           "TesT_ZERO_SUPPRESS_WORD_8_0_COMPLEX_8", \
                           i, start=7.2, end=100.1)
        print len(vect[0])
        print vect
    #--------------------------------------------------------------------
    # Test reading frames sequentially
    #--------------------------------------------------------------------
    frameIteration( "Z-R_std_test_frame_ver8-600000000-1.gwf" )

