#ifdef HAVE_CONFIG_H
#include <ldas_tools_config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <iomanip>

#include "ldastoolsal/unittest.h"

#include "genericAPI/StatFork.hh"
#include "genericAPI/StatPool.hh"
#include "genericAPI/Stat.hh"

using namespace GenericAPI;

LDASTools::Testing::UnitTest Test;

namespace
{
  const std::string	abs_builddir( ::getenv( "ABS_BUILDDIR" ) );

  const std::string lstat_command( abs_builddir + "/../src/lstat" );
  class TestStat
    : public StatFork
  {
  public:
    TestStat( );
  };

  TestStat::
  TestStat( )
  {
    SetCommand( lstat_command );
  }

  void
  lstat_pool_cb( StatPool::info_type Key )
  {
    StatFork*	flstat = NULL;

    if ( ( flstat =dynamic_cast< StatFork* >( Key ) ) )
    {
      flstat->SetCommand( lstat_command );
    }
  }

  bool
  ValidateStatBuf( const std::string& Element, bool Cond )
  {
    Test.Check( Cond )
      << Element
      << std::endl;
    return Cond;
  }

  void
  ValidateStat( const struct stat& Local, const struct stat& Remote,
		int LocalRetCode, int RemoteRetCode ,
		int LocalErrCode, int RemoteErrCode )
  {
    //-------------------------------------------------------------------
    // Verify that they have generated the same results
    //-------------------------------------------------------------------
    bool	cond = ( LocalRetCode == RemoteRetCode );

    Test.Check( cond ) << " return code equivalence"
		       << " (" << LocalRetCode << " vs. "
		       << RemoteRetCode << ")"
		       << std::endl;
    if ( cond )
    {
      if ( LocalRetCode == 0 )
      {
	//---------------------------------------------------------------
	// Both were successful with getting information.
	// Ensure the correctness of the data retrieved
	//---------------------------------------------------------------
	ValidateStatBuf( "st_dev", Local.st_dev == Remote.st_dev );
	ValidateStatBuf( "st_ino", Local.st_ino == Remote.st_ino );
	ValidateStatBuf( "st_mode", Local.st_mode == Remote.st_mode );
	ValidateStatBuf( "st_nlink", Local.st_nlink == Remote.st_nlink );
	ValidateStatBuf( "st_uid", Local.st_uid == Remote.st_uid );
	ValidateStatBuf( "st_gid", Local.st_gid == Remote.st_gid );
	ValidateStatBuf( "st_rdev", Local.st_rdev == Remote.st_rdev );
	ValidateStatBuf( "st_size", Local.st_size == Remote.st_size );
	ValidateStatBuf( "st_blksize", Local.st_blksize == Remote.st_blksize );
	if ( ValidateStatBuf( "st_blocks",
			      Local.st_blocks == Remote.st_blocks ) == false )
	{
	  Test.Message( ) << "  Expected: " << Local.st_blocks
			  << " Received: " << Remote.st_blocks
			  << std::endl;
	}
	ValidateStatBuf( "st_atime", Local.st_atime == Remote.st_atime );
	ValidateStatBuf( "st_mtime", Local.st_mtime == Remote.st_mtime );
	ValidateStatBuf( "st_ctime", Local.st_ctime == Remote.st_ctime );
      }
      else
      {
	//-------------------------------------------------------------
	// Both were unsuccessful with getting information
	// Ensure the correctness of the error code
	//-------------------------------------------------------------
	Test.Check( LocalErrCode == RemoteErrCode )
	  << " errno equivalence"
	  << " (" << LocalErrCode << " vs. "
	  << RemoteErrCode << ")"
	  << std::endl;
      }
    }
    else
    {
      Test.Message( ) << " Error codes: Local: " << LocalErrCode
		      << ": " << strerror( LocalErrCode )
		      << std::endl;
      Test.Message( ) << " Error codes: Remote: " << RemoteErrCode
		      << ": " << strerror( RemoteErrCode )
		      << std::endl;
    }
  }

  void
  Validate( const std::string& Filename )
  {
    enum {
      LOCAL = 0,
      REMOTE,
      POOL,
      FUNCTION,
      END_OF_LIST
    };

    int		retcode[ END_OF_LIST ];
    struct stat	statbuf[ END_OF_LIST ];
    int		errcode[ END_OF_LIST ];

    //-------------------------------------------------------------------
    // Run stat command both locally and remote
    //-------------------------------------------------------------------
    //...................................................................
    // Stat using local lstat
    //...................................................................
    retcode[ LOCAL ] = ::lstat( Filename.c_str( ), &( statbuf[ LOCAL ] ) );
    errcode[ LOCAL ] = errno;
    {
      //.................................................................
      // Stat using process
      //.................................................................
      TestStat	tlstat;

      tlstat.Spawn( );
      retcode[ REMOTE ] = tlstat.LStat( Filename, statbuf[ REMOTE ] );
      errcode[ REMOTE ] = errno;

      ValidateStat( statbuf[ LOCAL ], statbuf[ REMOTE ],
		    retcode[ LOCAL ], retcode[ REMOTE ],
		    errcode[ LOCAL ], errcode[ REMOTE ] );
    }
    {
      //.................................................................
      // Stat using POOL
      //.................................................................
      StatPool::UserInitCB( lstat_pool_cb );

      StatPool::info_type tlstat = StatPool::Request( );
      retcode[ POOL ] = tlstat->LStat( Filename, statbuf[ POOL ] );
      errcode[ POOL ] = errno;
      ValidateStat( statbuf[ LOCAL ], statbuf[ POOL ],
		    retcode[ LOCAL ], retcode[ POOL ],
		    errcode[ LOCAL ], errcode[ POOL ] );
      StatPool::Release( tlstat );
    }
    {
      //.................................................................
      // Stat using FUNCTION
      //.................................................................
      retcode[ FUNCTION ] = GenericAPI::LStat( Filename, statbuf[ FUNCTION ] );
      errcode[ FUNCTION ] = errno;
      ValidateStat( statbuf[ LOCAL ], statbuf[ FUNCTION ],
		    retcode[ LOCAL ], retcode[ FUNCTION ],
		    errcode[ LOCAL ], errcode[ FUNCTION ] );
    }
  }
}

int
main( int ArgC, char** ArgV )
{
  Test.Init(ArgC, ArgV);

  std::string	abs_builddir( ::getenv( "ABS_BUILDDIR" ) );
  std::string	filename;

  filename = abs_builddir + "/tlstat";
  Validate( filename );
  Validate( "./Makefile" );
  Validate( "./NoSuchFile.xyz" );

  Test.Exit( );

  return 1;	// Should never get here so exit with error if it happens
}
